from searchable_collection.simple_comparator import SimpleComparator
from searchable_collection.item import SearchableItem
from searchable_collection.collection import SearchableCollection
__version__ = "0.0.3"


from searchable_collection.item import SearchableItem





class SearchableCollection:
    def __init__(self,list_initial=[], master=None):
        self.master = master
        self.items = []
        self.extend(list_initial)

    def __getitem__(self, index):
        """
        allows us to index into our OrderList just like a normal array

        :param index: the integer index of the Order to access
        :return: the order at the specified index
        """
        if isinstance(index,slice):
            return [i.target for i in self.items[index]]
        return self.items[index].target

    def __setitem__(self, index, value):
        self.items[index] = SearchableItem(value,self)
        self._reindex()

    def __iter__(self):
        """
        allow orders to be iterated over
        :return: an iterator of all the orders
        """
        return (item.target for item in self.items[:])

    def __len__(self):
        return len(self.items)

    def _reindex(self):
        """
        reassigns indexes to all orders in self.orders, this should not be called directly in general

        :return:
        """
        for i, item in enumerate(self.items):
            item.set_item_index(i)

    def extend(self, other_list):
        for item in other_list:
            self.append_item(item, False)
        self._reindex()

    def pop(self, index=-1):
        item = self.items.pop(index)
        return item.target

    def insert(self, index, item, reindex=True):
        return self.insert_item(index, item, reindex)

    def append(self, item, reindex=True):
        return self.append_item(item, reindex)

    def insert_item(self, index, item, reindex=True):
        self.items.insert(index, SearchableItem(item, self))
        if reindex:
            self._reindex()
        return self.items[-1].target

    def append_item(self, item, reindex=True):
        self.items.append(SearchableItem(item, self))
        if reindex:
            self._reindex()
        return self.items[-1].target

    def delete(self, item):
        """
        delete a given order
        :param item: an integer index or an order instance
        :return:
        """
        if isinstance(item, int) and item not in self.items:
            del self.items[item]
        else:
            self.items.remove(item)
        self._reindex()

    def find_one_where(self, **kwargs):
        """

        order = orders.find_one_where(sn="123123")
        order = orders.find_one_where(sn="123123",reading_type="Fat")

        :param kwargs:Accepts same params as Order constructor (sn devType reading_type, etc)
        :return: the first matching order found or None
        """
        try:
            return next(self.find_all_where(**kwargs))
        except StopIteration:
            return None

    def find_all_where(self, **kwargs):
        """
        orders = list(orders.find_all_where(sn="123123"))
        for item in orders.find_all_where(sn="123123",reading_type="Fat"): ...

        :params : Accepts same params as Order constructor (sn devType reading_type, etc)
        :return: an iterator of all orders matching the provided criteria
        """
        def test_value(item,key,value):
            result = getattr(item, key, None)
            if callable(result):
                try:
                    result = result()
                except TypeError:
                    try:
                        return result(value)
                    except Exception as e:
                        return False
            return result == value
        for item in self.items[:]:
            if all(test_value(item,key,val) for key, val in kwargs.items()):
                yield item.target

    def delete_where(self, **kwargs):
        '''

        orders.delete_where(sn="123123")
        orders.delete_where(sn="123123",reading_type="Fat")

        :params : Accepts same params as Order constructor (sn devType reading_type, etc)
        :return: None
        '''
        self.items = [item for item in self.items if item not in list(self.find_all_where(**kwargs))]
        self._reindex()

    def to_dicts(self):
        '''
        convert to a list of dictionaries and return it
        :return: a list of dictionaries
        '''
        return [item.to_dict() for item in self.items]






from operator import eq
from searchable_collection.simple_comparator import SimpleComparator

class SearchableItem:
    collection = None
    _item_index = -1

    def __init__(self, target_ob, collection):
        self.collection = collection
        self.target = target_ob
        self.original_delete = getattr(target_ob, "delete", lambda *a: 1)

        try:
            self.target.collection = self.collection
            self.target.delete = lambda *a: SearchableItem.delete(self)
            self.target.item_index = self._item_index
        except:
            pass

    def set_item_index(self, val):
        self._item_index = val
        try:
            self.target.item_index = val
        except:
            pass

    def delete(self):
        self.collection.delete(self)

    def __eq__(self, other):
        return eq(self.target, other)

    def __getattr__(self, item):
        if hasattr(self.target, item):
            return getattr(self.target, item)
        elif isinstance(self.target,dict) and item in self.target:
            return self.target[item]
        elif item == "to_dict":
            return lambda: getattr(self.target, '__dict__', self.target)
        elif item in SimpleComparator.comparator_dict:
            return SimpleComparator(self.target, item)
        elif item.startswith("not_") and item[4:] in SimpleComparator.comparator_dict:
            return SimpleComparator(self.target, item[4:],negate=True)
        try:
            rest, end_part = item.split("__", 1)
        except:
            # print("CANNOT SPLIT ITEM %r???" % item)
            print("No Attribute %r on %r" % (item,self.target))
            return None


        target = result = self
        while hasattr(result, rest) or isinstance(result,dict) and rest in result:
            target = result
            if isinstance(target,dict):
                result = target[rest]
            else:
                result = getattr(target, rest)
            try:
                rest, end_part = end_part.split("__", 1)
            except:
                break

        if end_part in SimpleComparator.comparator_dict:
            return SimpleComparator(result, end_part)
        elif end_part.startswith("not_") and end_part[4:] in SimpleComparator.comparator_dict:
            return SimpleComparator(result, end_part[4:], negate=True)
        elif isinstance(result,dict):
            return result.get(end_part,result)
        return getattr(result, end_part, result)




from setuptools import setup
from searchable_collection import __version__
setup(
    name='searchable_collection',
    version=__version__,
    packages=['searchable_collection'],
    url='http://searchablecollection.readthedocs.io/en/latest/',
    license='MIT',
    author='Joran Beasley',
    author_email='joranbeasley@gmail.com',
    repo='https://github.com/joranbeasley/searchable_collection',
    description='A searchable list implementation, written entirely in python.',
    classifiers=[
          'Development Status :: 4 - Beta',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: Microsoft :: Windows',
          'Operating System :: POSIX',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development',
          'Topic :: Software Development :: Bug Tracking',
          'Topic :: Software Development :: Libraries',
          'Topic :: Database',
          'Topic :: Utilities',
          ]
)



import re
from operator import lt, contains, gt, ge, le, eq


def startswith(x, y):
    return x.startswith(y)


def endswith(x, y):
    return x.endswith(y)


def is_in(x, y):
    """
    effectively operator.in

    :param x: needle
    :param y: haystack
    :return: x in y
    """
    return x in y


def re_match(x, y, flags=0):
    """
    effectively operator.re_match

    :param x: search_text
    :param y: pattern
    :param flags: re.I|re.DOTALL|... see re documentation
    :return: re.match(pattern,search_text,flags)
    """
    return re.match(y, x, flags)


def re_search(x, y, flags=0):
    """
    effectively operator.re_search

    :param x: search_text
    :param y: pattern
    :param flags: re.I|re.DOTALL|... see re documentation
    :return: re.search(pattern,search_text,flags)
    """
    return re.search(y, x, flags)


class SimpleComparator:
    comparator_dict = {
        # numeric (mostly)
        'lt': lt, 'gt': gt, "lte": le, 'gte': ge, 'ge': ge, 'le': le, 'eq': eq,
        # string matching
        'contains': contains, "in": is_in, 'startswith': startswith, "endswith": endswith,
        # regular expressions
        "match": re_match, "search": re_search,
        # helper to not shadow in if its the only part of the argument .find_where(in=[1,2,3]) =>.find_where(is_in=[1,2,3])
        'is_in': is_in
    }

    def __init__(self, item, comparator, negate=False):
        self.item = item
        self.comparator = comparator
        self.comparator_fn = self.comparator_dict[comparator]
        self.negate = negate

    def __eq__(self, other):
        try:
            if self.negate:
                return not self.comparator_fn(self.item, other)
            return self.comparator_fn(self.item, other)
        except (ValueError, TypeError, AttributeError):
            return self.negate

